<?php

include_once __DIR__ . '/../vendor/autoload.php';

use GuzzleHttp\Client;

/**
 * Class CareCloudAPI
 */
class CareCloudAPI
{
  public $client = null;

  private CareCloudCustomer $CareCloud;

  private $token = null;

  private $customer_id = null;

  private $errors = [];

  public function __construct(CareCloudCustomer $CareCloud)
  {

    $this->CareCloud = $CareCloud;

    // Do not initiate when prop not set
    if (empty($this->CareCloud->getApiURL()))
      return;

    $this->client = new GuzzleHttp\Client([
      'base_uri' => $this->getAPIURL(),
      'timeout' => 15.0,
    ]);
  }

  public function APIHasSetBasicVars(bool $flush = false): bool
  {
    return !(empty($this->CareCloud->getApiURL($flush)) || empty($this->CareCloud->getExternalAppID($flush)));
  }

  public function setSessionToken($token, bool $force = false)
  {
    if (empty($_SESSION['crmcarecloud_api_token']) || $force === true) {
      $_SESSION['crmcarecloud_api_token'] = $token;
    }

    $this->token = $_SESSION['crmcarecloud_api_token'];

    return $this->token;
  }

  public function getSessionToken($flush = false)
  {
    if ($this->token === null || $flush === true)
      $this->obtainSessionToken($flush);

    return $this->token;
  }

  private function hasSessionToken()
  {
    return $this->token !== null;
  }

  private function obtainSessionToken(bool $flush = false)
  {
    if ($this->token === null || $flush === true) {
      if (empty($_SESSION['crmcarecloud_api_token']) || $flush === true)
        $this->customerAPITokenController();
      else
        $this->token = $_SESSION['crmcarecloud_api_token'];
    }
  }

  public function flushSessionToken()
  {
    unset($_SESSION['crmcarecloud_api_token']);
  }

  public function getAPIURL()
  {
    return $this->CareCloud->getApiURL();
  }

  public function setCustomerID($id)
  {
    $_SESSION['carecloud_customer_id'] = $this->customer_id = $id;

    return $this->customer_id;
  }

  public function getCustomerID()
  {
    if ($this->customer_id === null)
      $this->obtainCustomerID();

    return $this->customer_id;
  }

  public function hasCustomerID()
  {
    return $this->customer_id !== null;
  }

  public function obtainCustomerID() : void
  {
    if (empty($_SESSION['carecloud_customer_id'])) {
      // Redirect to login page
    } else {
      $this->customer_id = $_SESSION['carecloud_customer_id'];
    }
  }

  public function isCustomerTestUser()
  {
    return !!$_SESSION['carecloud_customer_is_test_user'];
  }

  public function hasErrors()
  {
    return (count($this->errors) > 0);
  }

  public function getErrors()
  {
    return $this->errors;
  }

  public function getMessagesAsArrayForVue()
  {
    $res = [];

    if ($this->hasErrors()) {
      $errs = $this->getErrors();
      foreach ($errs as $err) {
        $res[] = array(
          "type" => "error",
          "content" => $err['title']
        );
      }
    }

    return $res;
  }

  public function setError(int $errorCode, string $title, string $message)
  {
    array_push($this->errors, array('code' => $errorCode, 'title' => $title, 'message' => $message));
  }

  public function printFormattedErrors()
  {
    $errs = $this->getErrors();
    $error = "";

    foreach ($errs as $err) {
      $error .= '<div class="notice notice-error is-dismissible"><p class="form-error">' . $err['title'] . '</p></div>';
    }

    // Empty errors
    $this->errors = [];

    return $error;
  }

  public function customerAPITokenController()
  {

    $response = $this->client->request('POST', 'tokens', [
      'headers' => [
        'Authorization' => 'Basic Y3VzdG9tZXJfaW50ZXJmYWNlOg==',
        'Content-Type' => 'application/json',
        'Accept-Language' => $this->CareCloud->getLanguage(),
      ],
      "body" => json_encode([
        "device" => [
          "device_id" => session_id(),
          "device_system" => "1",
          "device_name" => "1",
          "device_type" => "1",
        ],
        "setup" => [
          "language_id" => "en",
          "allowed_gps" => "false",
          "allowed_notifications" => "false",
        ],
        "external_application_id" => $this->CareCloud->getExternalAppID(),
        "push_token" => "push_token_test",
      ]),
    ]);

    if ($response->getBody()) {
      $token = json_decode($response->getBody()->getContents());
      $token = $token->data->token_id;
      $this->setSessionToken($token, true);
    }
  }

  public function customerAPILoginController()
  {

    try {
      $authorization_token = base64_encode('customer_interface:' . $this->getSessionToken());
      $response = $this->client->request('POST', 'tokens/' . $this->getSessionToken() . '/actions/login', array(
        'headers' => array(
          'Authorization' => 'Basic ' . $authorization_token,
          'Content-Type' => 'application/json',
          'Accept-Language' => $this->CareCloud->getLanguage(),
        ),
        'body' => json_encode(array(
          'login_type' => 'email',
          'login_value' => trim($_POST['carecloud-api-login-mail']),
          'password' => $_POST['carecloud-api-login-password'],
        )),
      ));
    } catch (\GuzzleHttp\Exception\ClientException $e) {

      $res = json_decode($e->getResponse()->getBody()->getContents());

      if ($res->error->exception === "invalid_credentials_exception")
        $this->setError($e->getCode(), $this->CareCloud->getTranslation('invalid_login_credentials'), '');
      elseif ($res->error->exception === "bad_request_exception")
        $this->setError($e->getCode(), $this->CareCloud->getTranslation('something_went_wrong'), '');
      else
        $this->setError($e->getCode(), $res->error->title, '');
    } catch (\GuzzleHttp\Exception\ConnectException $e) {
      $this->setError($e->getCode(), "Could not connect to CRM CareCloud.", 'Please contact support.');
    } catch (\Exception $e) {
      $this->setError("400", "Unknown error.");
    }

    if (!$this->hasErrors()) {
      $result = json_decode($response->getBody()->getContents());
      $this->setCustomerID($result->data->customer_id);
      $_SESSION['crmcarecloud_new_log_in'] = true;
      $this->CareCloud->redirectToCustomerOverviewPage('header');
    }
  }

  public function isAfterLoginAccess(): bool
  {

    if (isset($_SESSION['crmcarecloud_new_log_in']) && $_SESSION['crmcarecloud_new_log_in'] == 1) {
      unset($_SESSION['crmcarecloud_new_log_in']);
      return true;
    } else
      return false;

  }

  public function loginTestUser()
  {
    $response = null;

    try {
      $response = $this->customerAPICall('tokens/' . $this->getSessionToken() . '/actions/login-test-customer', 'POST');
    } catch (\GuzzleHttp\Exception\BadResponseException $e) {
      $res = json_decode($e->getResponse()->getBody()->getContents());
      $this->setError($e->getCode(), $res->error->title, '');
    } catch (\Exception $e) {
      $this->setError("400", "Unknown error.");
    }

    if ($response !== null && property_exists($response, 'customer_id')) {
      $_SESSION['carecloud_customer_is_test_user'] = true;
      $this->setCustomerID($response->customer_id);
      $customer = $this->customerAPIUserCall('');
      $_SESSION['carecloud_test_customer'] = array(
        'first_name' => $customer->personal_information->first_name,
        'last_name' => $customer->personal_information->last_name,
      );
    }

    if ($response !== null && property_exists($response, 'error')) {
      $this->setError(400, $response->error->title, '');
    }

    return $response;
  }

  public function logoutController($redirect = true)
  {
    if ($this->hasSessionToken() && !$this->isCustomerLogged()) {
      return null;
    }

    $authorization_token = base64_encode('customer_interface:' . $this->getSessionToken());

    try {
      $response = $this->client->request('POST', 'tokens/' . $this->getSessionToken() . '/actions/logout', array(
        'headers' => array(
          'Authorization' => 'Basic ' . $authorization_token,
          'Content-Type' => 'application/json',
          'Accept-Language' => $this->CareCloud->getLanguage(),
        ),
        'body' => json_encode(array(
          'token_id' => $this->getSessionToken(),
        )),
      ));
    } catch (\GuzzleHttp\Exception\ClientException $e) {

      if (wp_redirect(get_page_link($this->CareCloud->getLoginPage()))) {
        exit;
      }
    }


    $_SESSION['crmcarecloud_api_token'] = null;
    $_SESSION['carecloud_customer_id'] = null;

    if ($redirect) {
      $this->CareCloud->redirectToLoginPage('header');
      exit;
    }

    return $response;
  }

  public function customerAPICall($route, $method = "GET", $jsonData = array())
  {

    $authorization_token = base64_encode('customer_interface:' . $this->getSessionToken());

    if ($route === 'send-password-setup-email') {
      $route = "tokens/" . $this->getSessionToken() . "/actions/send-password-setup-email";
    }

    if ($route === 'temporary-login') {
      $route = "tokens/" . $this->getSessionToken() . "/actions/login";
    }

    if ($route === 'temporary-logout') {
      $route = "tokens/" . $this->getSessionToken() . "/actions/logout";
    }


    $data = array(
      'headers' => array(
        'Authorization' => 'Basic ' . $authorization_token,
        'Content-Type' => 'application/json',
        'Accept-Language' => $this->CareCloud->getLanguage(),
      ),
    );

    if (!empty($jsonData)) {
      array_walk_recursive($jsonData, function (&$item, $key) {
        $item = ($item === "") ? null : $item;
      });
      $data = array_merge($data, array('json' => $jsonData));
    }

    $response = $this->client->request($method, $route, $data);

    $result = json_decode($response->getBody()->getContents());

    return $result->data;
  }

  public function isCustomerLogged(): bool
  {
    return ($this->getCustomerID() !== null);
  }

  public function customerAPIUserCall($route, $method = "GET", $data = array())
  {
    return $this->customerAPICall('customers/' . $this->getCustomerID() . "/" . $route, $method, $data);
  }
}